'''
Created on Feb 18, 2017

@author: julien
'''
import numpy

from keras import activations, backend
from keras.datasets import reuters
from keras.engine.topology import Layer
from keras.preprocessing.text import Tokenizer
from keras.utils import np_utils

from minos.train.utils import SimpleBatchIterator


def get_reuters_dataset(batch_size, max_words):
    (X_train, y_train), (X_test, y_test) = reuters.load_data(nb_words=max_words, test_split=0.2)

    tokenizer = Tokenizer(nb_words=max_words)
    X_train = tokenizer.sequences_to_matrix(X_train, mode='binary')
    X_test = tokenizer.sequences_to_matrix(X_test, mode='binary')

    nb_classes = numpy.max(y_train) + 1
    y_train = np_utils.to_categorical(y_train, nb_classes)
    y_test = np_utils.to_categorical(y_test, nb_classes)

    batch_iterator = SimpleBatchIterator(
        X_train,
        y_train,
        batch_size,
        autoloop=True)
    test_batch_iterator = SimpleBatchIterator(
        X_test,
        y_test,
        len(X_test),
        autoloop=True)
    return batch_iterator, test_batch_iterator, nb_classes


class CustomLayer(Layer):

    def __init__(self, output_dim=None, activation=None, **kwargs):
        self.output_dim = output_dim
        self.activation = activation
        super().__init__(**kwargs)

    def build(self, input_shape):
        self.W = self.add_weight(
            shape=(input_shape[1], self.output_dim),
            initializer='glorot_uniform',
            trainable=True)
        self.b = self.add_weight(
            shape=(self.output_dim,),
            initializer='glorot_uniform',
            trainable=True)
        super().build(input_shape)  # Be sure to call this somewhere!

    def call(self, x, mask=None):
        activation = activations.get(self.activation)
        return activation(backend.dot(x, self.W) + self.b)

    def get_output_shape_for(self, input_shape):
        return (input_shape[0], self.output_dim)

    def get_config(self):
        config = {
            'output_dim': self.output_dim,
            'activation': self.activation}
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))



'''
Created on Mar 7, 2017

@author: julien
'''
import logging


def cpu_device():
    return '/cpu:0'


def get_available_gpus():
    try:
        from tensorflow.python.client import device_lib
        local_device_protos = device_lib.list_local_devices()
        return [x.name for x in local_device_protos if x.device_type == 'GPU']
    except Exception as ex:
        logging.error(
            'Error while trying to list available GPUs: %s' % str(ex))
        return list()


def default_device():
    gpus = get_available_gpus()
    if len(gpus) > 0:
        return gpus[1]
    return cpu_device()


def get_logical_device(physical_device):
    if is_gpu_device(physical_device):
        return '/gpu:0'
    return physical_device


def is_cpu_device(device):
    return device\
        and isinstance(device, str)\
        and device.startswith('/cpu')


def is_gpu_device(device):
    return device\
        and isinstance(device, str)\
        and device.startswith('/gpu')


def get_device_idx(device):
    return device.split(':')[1]


def setup_tf_session(device):
    import tensorflow as tf
    config = tf.ConfigProto()
    if hasattr(config, 'gpu_options'):
        config.allow_soft_placement = True
        config.gpu_options.allow_growth = True
    if is_gpu_device(device):
        config.gpu_options.visible_device_list = str(
            get_device_idx(device))
    elif is_cpu_device(device):
        config.gpu_options.visible_device_list = ''
    from keras import backend
    session = tf.Session(config=config)
    session.run(tf.global_variables_initializer())
    backend.set_session(session)



'''
Created on Feb 6, 2017

@author: julien
'''
import numpy
from os.path import join
import tempfile


from examples.ga.dataset import get_reuters_dataset
from minos.experiment.experiment import Experiment, ExperimentParameters
from minos.experiment.training import Training, AccuracyDecreaseStoppingCondition,\
    get_associated_validation_metric
from minos.model.design import create_random_blueprint
from minos.model.model import Objective, Optimizer, Metric, Layout
from minos.train.trainer import ModelTrainer
from minos.utils import setup_console_logging
import numpy as np


np.random.seed(1337)


def create_experiment(input_size, output_size, batch_size):
    training = Training(
        objective=Objective('categorical_crossentropy'),
        optimizer=Optimizer(optimizer='Adam'),
        metric=Metric('categorical_accuracy'),
        stopping=AccuracyDecreaseStoppingCondition(
            metric='categorical_accuracy',
            min_epoch=5,
            max_epoch=25,
            noprogress_count=5),
        batch_size=batch_size)
    parameters = ExperimentParameters(use_default_values=True)
    layout = Layout(
        input_size=input_size,
        output_size=output_size,
        output_activation='softmax')
    experiment = Experiment(
        label='reuters_train_multi_gpu',
        layout=layout,
        training=training,
        parameters=parameters)
    return experiment

def train_multi_gpu(max_words = 1000, batch_size=32):
    batch_iterator, test_batch_iterator, nb_classes = get_reuters_dataset(batch_size, max_words)
    experiment = create_experiment(max_words, nb_classes, batch_size)
    blueprint = create_random_blueprint(experiment)
    devices = ['/gpu:0', '/gpu:1']
    trainer = ModelTrainer(batch_iterator, test_batch_iterator)
    with tempfile.TemporaryDirectory() as tmp_dir:
        model, history, _duration = trainer.train(
            blueprint,
            devices,
            save_best_model=True,
            model_filename=join(tmp_dir, 'model'))
        metric = get_associated_validation_metric(blueprint.training.metric.metric)
        epoch = numpy.argmax(history.history[metric])
        score = history.history[metric][epoch]
        print('Final training score %r after %d epoch' % (score, epoch))
        
        test_size = 10
        y_true = numpy.argmax(test_batch_iterator.y[0][:test_size], axis=1)
        y_pred = numpy.argmax(model.predict(test_batch_iterator.X[0][:test_size]), axis=1)
        evaluation = numpy.mean(y_true==y_pred)
        print('Final evaluation score %f' % evaluation)
        print('Predictions (true, pred) %r' % list(zip(y_true.tolist(), y_pred.tolist())))
        
def main():
    setup_console_logging('DEBUG')
    train_multi_gpu()

if __name__ == '__main__':
    main()



'''
Created on Feb 12, 2017

@author: julien
'''
import logging
from multiprocessing import Queue, Process
from threading import Thread
from time import time
import traceback

from keras.callbacks import ModelCheckpoint
import numpy

from minos.experiment.training import EpochStoppingCondition,\
    AccuracyDecreaseStoppingCondition, AccuracyDecreaseStoppingConditionWrapper,\
    get_associated_validation_metric
from minos.tf_utils import setup_tf_session
from minos.utils import disable_sysout, load_keras_model


class MultiProcessModelTrainer(object):

    def __init__(self, batch_iterator, test_batch_iterator, environment):
        self.batch_iterator = batch_iterator
        self.test_batch_iterator = test_batch_iterator
        self.environment = environment

    def build_and_train_models(self, blueprints):
        logging.debug('Training %d models' % len(blueprints))
        return self._start_training_workers(blueprints)

    def _start_training_workers(self, blueprints):
        try:
            total_n_jobs = sum(self.environment.n_jobs)
            work_queue = Queue(total_n_jobs)
            result_queue = Queue(total_n_jobs)
            self.processes = [
                Process(
                    target=model_training_worker,
                    args=(
                        self.batch_iterator,
                        self.test_batch_iterator,
                        device_id,
                        device,
                        work_queue,
                        result_queue))
                for device_id, device in enumerate(self.environment.devices)
                for _job in range(self.environment.n_jobs[device_id])]
            self.process_count = 0
            for process in self.processes:
                self.process_count += 1
                process.start()

            def _work_feeder():
                count = len(blueprints)
                for i, blueprint in enumerate(blueprints):
                    work_queue.put((i, count, blueprint))
                for _ in range(sum(self.environment.n_jobs)):
                    work_queue.put(None)
            Thread(target=_work_feeder).start()

            results = []
            while self.process_count > 0:
                result = result_queue.get()
                if result:
                    logging.debug(
                        'Blueprint %d: best score %f, epoch %d/%d',
                        result[0],
                        result[1],
                        result[2],
                        result[3])
                    results.append(result)
                else:
                    self.process_count -= 1
            results = list(
                sorted(
                    results,
                    key=lambda e: e[0]))
            return results
        except Exception as ex:
            logging.error(ex)
            logging.error(traceback.format_exc())


class ModelTrainer(object):

    def __init__(self, batch_iterator, test_batch_iterator):
        from minos.model.build import ModelBuilder
        self.model_builder = ModelBuilder()
        self.batch_iterator = batch_iterator
        self.test_batch_iterator = test_batch_iterator

    def train(self, blueprint, device,
              save_best_model=False, model_filename=None, class_weight=None):
        try:
            model = self.model_builder.build(
                blueprint,
                device)
            class_weight = class_weight or blueprint.training.class_weight
            setup_tf_session(device)
            nb_epoch, callbacks = self._get_stopping_parameters(blueprint)
            if save_best_model:
                callbacks.append(self._get_model_save_callback(
                    model_filename,
                    blueprint.training.metric.metric))
            start = time()
            history = model.fit_generator(
                self.batch_iterator,
                self.batch_iterator.samples_per_epoch,
                nb_epoch,
                callbacks=callbacks,
                validation_data=self.test_batch_iterator,
                nb_val_samples=self.test_batch_iterator.sample_count,
                class_weight=class_weight)
            if save_best_model:
                del model
                model = load_keras_model(model_filename)
            return model, history, (time() - start)
        except Exception as ex:
            logging.debug(ex)
            logging.debug(traceback.format_exc())
        try:
            from keras import backend
            backend.clear_session()
        except:
            logging.debug(ex)
            logging.debug(traceback.format_exc())
        return None, None, 0

    def _get_model_save_callback(self, model_filename, metric):
        checkpoint = ModelCheckpoint(
            model_filename,
            monitor=metric,
            save_best_only=True)
        return checkpoint

    def _get_stopping_parameters(self, blueprint):
        if isinstance(blueprint.training.stopping, EpochStoppingCondition):
            nb_epoch = blueprint.training.stopping.epoch
            stopping_callbacks = []
        if isinstance(blueprint.training.stopping, AccuracyDecreaseStoppingCondition):
            nb_epoch = max(
                1,
                blueprint.training.stopping.min_epoch,
                blueprint.training.stopping.max_epoch)
            stopping_callbacks = [
                AccuracyDecreaseStoppingConditionWrapper(blueprint.training.stopping)]
        return nb_epoch, stopping_callbacks


def model_training_worker(batch_iterator, test_batch_iterator,
                          device_id, device, work_queue, result_queue):
    disable_sysout()
    model_trainer = ModelTrainer(
        batch_iterator,
        test_batch_iterator)
    work = work_queue.get()
    while work:
        try:
            idx, _total, blueprint = work
            model, history, duration = model_trainer.train(
                blueprint,
                device)
            if model and history:
                epoch_total = len(history.epoch)
                val_metric = get_associated_validation_metric(blueprint.training.metric.metric)
                epoch_best = numpy.argmax(history.history[val_metric])
                score = history.history[val_metric][epoch_best]
            else:
                score, epoch_best, epoch_total = 0, 0, 0
            result_queue.put((idx, score, epoch_best, epoch_total, blueprint, duration, device_id))
            work = work_queue.get()
        except Exception as ex:
            logging.error(ex)
    result_queue.put(None)



'''
Created on Feb 18, 2017

@author: julien
'''
from os.path import join
import tempfile
import unittest

from minos.experiment.experiment import ExperimentParameters, Experiment
from minos.experiment.training import Training, EpochStoppingCondition
from minos.model.build import ModelBuilder
from minos.model.design import create_random_blueprint
from minos.model.model import Layout, Objective, Optimizer, Metric
from minos.model.parameter import int_param
from minos.model.parameters import register_custom_activation,\
    register_custom_layer
from minos.train.utils import default_device, CpuEnvironment
from minos.utils import disable_sysout, load_model, load_keras_model
from tests.fixtures import get_reuters_dataset, CustomLayer


class ModelSaveTest(unittest.TestCase):

    def test_save(self):
        disable_sysout()

        def custom_activation(x):
            return x

        register_custom_activation('custom_activation', custom_activation)
        register_custom_layer('custom_layer', CustomLayer, {'output_dim': int_param(1, 100)})

        with tempfile.TemporaryDirectory() as tmp_dir:
            batch_size = 50
            batch_iterator, test_batch_iterator, nb_classes = get_reuters_dataset(batch_size, 1000)
            layout = Layout(
                input_size=1000,
                output_size=nb_classes,
                output_activation='softmax')
            training = Training(
                objective=Objective('categorical_crossentropy'),
                optimizer=Optimizer(optimizer='Adam'),
                metric=Metric('categorical_accuracy'),
                stopping=EpochStoppingCondition(10),
                batch_size=batch_size)
            experiment_parameters = ExperimentParameters(use_default_values=True)
            experiment_parameters.layout_parameter('rows', 1)
            experiment_parameters.layout_parameter('blocks', 1)
            experiment_parameters.layout_parameter('layers', 1)
            experiment_parameters.layout_parameter('block.layer_type', 'custom_layer')
            experiment = Experiment(
                'test__reuters_experiment',
                layout,
                training,
                batch_iterator,
                test_batch_iterator,
                CpuEnvironment(n_jobs=1, data_dir=tmp_dir),
                parameters=experiment_parameters)

            blueprint = create_random_blueprint(experiment)
            model = ModelBuilder().build(blueprint, default_device())
            model.fit_generator(
                generator=batch_iterator,
                samples_per_epoch=batch_iterator.samples_per_epoch,
                nb_epoch=10,
                validation_data=test_batch_iterator,
                nb_val_samples=test_batch_iterator.sample_count)
            filepath = join(tmp_dir, 'model')
            model.save(filepath)
            model = load_keras_model(filepath)
            self.assertIsNotNone(model, 'Should have loaded the model')


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()
